#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import sys
import json
import queue
import threading
import grpc

from .rpc import reverseproxy_pb2
from .rpc import reverseproxy_pb2_grpc

class MessageQueue:
    def __init__(self):
        self.q = queue.Queue()
    
    def pop(self):
        try:
          return self.q.get(True)
        except queue.Empty:
          return None

    def push(self, item):
        if self.q.full():
            self.q.get_nowait()
        self.q.put(item)


class ProxyAgent():
    def __init__(self, method, endpoint, handler):
        self.method = method
        self.endpoint = endpoint
        self.handler = handler
        self._mq = MessageQueue()
        self.__connect_to_server()

    def __connect_to_server(self):
        try:
            self._channel = grpc.insecure_channel(target='unix:/host/run/tpfunc/proxy.sock')
            self._conn = reverseproxy_pb2_grpc.TriggerStub(self._channel)

            profile = reverseproxy_pb2.Profile(method=self.method, endpoint=self.endpoint)
            confirm = self._conn.Register(profile)
            if confirm is not None:
                threading.Thread(target=self.__listen_for_message, daemon=True).start()
            else:
                raise Exception('register http profile failed')
        except Exception as e:
            print(e)

    def __construct_request(self, message):
        try:
            if len(message) > 0:
                return json.loads(message)
        except Exception as e:
            return message

    def __wait_response(self, method, endpoint, queue):
        # notify server to know http service start
        profile = reverseproxy_pb2.Profile(method=method, endpoint=endpoint)
        reply = reverseproxy_pb2.Reply(profile=profile)
        yield reply

        # wait blocking for response
        while True:
            yield queue.pop()

    def __construct_response(self, resp):
        data = ""
        try:
            if isinstance(resp.data, dict):
                data = json.dumps(resp.data)
            elif isinstance(resp.data, str):
                data = resp.data
            else:
                raise Exception('response data type either string or dict')
            return self.__empty_response(code=resp.code, payload=data)
        except Exception as e:
            return self.__empty_response(code=400, payload="{{\"message\":\"{}\"}}".format(e))

    def __empty_response(self, code=200, headers="{}", payload="{}"):
        return reverseproxy_pb2.Reply(code=code, headers=headers, payload=payload)

    def __listen_for_message(self):
        try:
            response_iterator = self.__wait_response(self.method, self.endpoint, self._mq)
            requests = self._conn.Fire(response_iterator)
            for request in requests:
                resp = self.handler(headers=request.headers, message=request.payload)
                reply = self.__empty_response() if resp is None else self.__construct_response(resp)
                self._mq.push(reply)
        except Exception as e:
            print(e)


class Response():
    def __init__(self, code=200, data={}):
        """ Constructor """
        self.code = code
        self.data = data


def http_get(resource=''):
    def decorator(func):
        def proxy_get():
            ProxyAgent(method='GET', endpoint=resource, handler=func)
        return proxy_get
    return decorator


def http_put(func, resource=''):
    def proxy_put():
        ProxyAgent('PUT', resource, func)
    return proxy_put


def http_post(func, resource=''):
    def proxy_post():
        ProxyAgent('POST', resource, func)
    return proxy_post


def http_delete(func, resource=''):
    def proxy_delete():
        ProxyAgent('DELETE', resource, func)
    return proxy_delete


class Server():
    @staticmethod
    def GET(resource='', handler=None):
        ProxyAgent('GET', resource, handler)

    @staticmethod
    def PUT(resource='', handler=None):
        ProxyAgent('PUT', resource, handler)

    @staticmethod
    def POST(resource='', handler=None):
        ProxyAgent('POST', resource, handler)

    @staticmethod
    def DELETE(resource='', handler=None):
        ProxyAgent('DELETE', resource, handler)




#!/usr/bin/python

import os
import json
import queue
import requests
import requests_unixsocket
import threading


class Token():
    def __init__(self):
        self._ip = os.getenv('APPMAN_HOST_IP', default='localhost')
        self._port = os.getenv('APPMAN_HOST_PORT', default=59000)
        self._prefix = os.getenv('MX_API_VER', default='api/v1/')

        with open("/var/run/mx-api-token") as f:
            token = f.readline()
            self._headers = {"Content-Type": "application/json", "mx-api-token": token.strip()}

class Access(Token):
    def __init__(self):
        super().__init__()
        self.__read_tag_url = "http://{}:{}/{}tags/access".format(self._ip, self._port, self._prefix)
        self.__write_tag_url = "http://{}:{}/{}tags/access".format(self._ip, self._port, self._prefix)

    def read(self, provider, source, tag):
        """ direct-read tag function """
        url = "{}/{}/{}/{}".format(self.__read_tag_url, provider, source, tag)
        r = requests.get(url, headers=self._headers, timeout=3)
        if r.status_code is 200:
            return r.status_code, r.json()
        else:
            return r.status_code, {}

    def write(self, provider, source, tag, data_type, data_value):
        """ direct-write tag function """
        url = "{}/{}/{}/{}".format(self.__write_tag_url, provider, source, tag)
        r = requests.put(url, headers=self._headers, json={"dataType": data_type, "dataValue": data_value}, timeout=3)
        if r.status_code is 200:
            return r.status_code, r.json()
        else:
            return r.status_code, {}

class Publisher(Token):
    def __init__(self):
        super().__init__()
        self._session = requests_unixsocket.Session()
        self.__pub_url = "http+unix://%2Frun%2Ftaghub%2Fhttp.sock/tags/publish"

    def publish(self, data):
        """ tag publish function """
        self._session.post(url=self.__pub_url, data=json.dumps(data))


class Subscriber(Token):
    def __init__(self):
        super().__init__()
        self._threads = {}
        self._session = requests_unixsocket.Session()
        self.__pub_url = "http+unix://%2Frun%2Ftaghub%2Fhttp.sock/tags/monitor"

    def __has_subscribed(self, provider, source):
        key = "{}/{}".format(provider, source)
        if self._threads.get(key, None) is not None:
            return True, key
        else:
            return False, ""

    def __get_endpoint(self, provider, source, tags):
        sep = ","
        url = "{}/{}/{}?tags={}&onChanged".format(self.__sub_url, provider, source, sep.join(tags))
        return url

    def __tag_monitoring(self, url):
        try:
            r = self._session.get(url=url, headers=self._headers, stream=True)

            if r.encoding is None:
                r.encoding = 'utf-8'

            for line in r.iter_lines(decode_unicode=True):
                if line:
                    data = json.loads(line[5:])
                    if hasattr(self, 'callback'):
                        self.callback(data)
        except Exception as e:
            print(e)

    def subscribe_callback(self, callback):
        self.callback = callback

    def subscribe(self, provider, source, tags):
        """ tag subscribe function (streaming) """
        if provider == "" or source == "" or not isinstance(tags, list):
            raise ValueError

        if len(tags) == 0:
            raise ValueError

        subscribed, topic = self.__has_subscribed(provider, source)
        if subscribed:
            return

        url = self.__get_endpoint(provider, source, tags)
        self._threads[topic] = threading.Thread(target=self.__tag_monitoring, args=(url, ), daemon=True).start()

    def unsubscribe(self, provider, source):
        if hasattr(self, '_threads'):
            subscribed, topic = self.__has_subscribed(provider, source)
            if not subscribed:
                return

            thread = self._threads.get(topic, None)
            if thread is not None:
                thread.stop()
                self._threads.pop(topic, None)




#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import time
import json
import requests
import threading


class Token():
    def __init__(self):
        self._ip = os.getenv('APPMAN_HOST_IP', default='localhost')
        self._port = os.getenv('APPMAN_HOST_PORT', default=59000)
        self._prefix = os.getenv('MX_API_VER', default='api/v1/')

        with open("/var/run/mx-api-token") as f:
            token = f.readline()
            self._headers = {"Content-Type": "application/json", "mx-api-token": token.strip()}


class Listener(Token):
    def __init__(self, callback):
        super().__init__()
        self.callback = callback
        self.__config = Configuration()
        self.__tag_url = "http://{}:{}/{}tags/monitor".format(self._ip, self._port, self._prefix)
        self.__event_url = "http://{}:{}/{}events".format(self._ip, self._port, self._prefix)

    def __get_tag_endpoint(self):
        urls = []
        tag_list = self.__config.data_driven_tags()
        for provider, sources in tag_list.items():
            for source, tags in sources.items():
                sep = ','
                url = "{}/{}/{}?tags={}&onChanged".format(self.__tag_url, provider, source, sep.join(tags))
                urls.append(url)
        return urls

    def __get_event_endpoint(self):
        sep = ','
        names = []
        categories = []
        event_list = self.__config.data_driven_events()
        for category, ns in event_list.items():
            categories.append(category)
            names = names + ns
        if not (len(categories) and len(names)):
            return ''
        return "{}?categories={}&eventNames={}&event=true".format(self.__event_url, sep.join(categories), sep.join(names))

    def __start_monitoring(self, url):
        while True:
            try:
                    r = requests.get(url, headers=self._headers, stream=True)

                    if r.encoding is None:
                        r.encoding = 'utf-8'

                    for line in r.iter_lines(decode_unicode=True):
                        if line:
                            data = json.loads(line[5:])
                            if hasattr(self, 'callback'):
                                self.callback(data)

            except Exception as e:
                print(e)
            time.sleep(1)

    def listen(self):
        """ subscribe function (streaming) """
        urls = self.__get_tag_endpoint()
        event_url = self.__get_event_endpoint()
        if event_url != '':
            urls.append(event_url)
        for url in urls:
            threading.Thread(target=self.__start_monitoring, args=(url, ), daemon=True).start()


class Configuration():
    def __init__(self):
        self.__load_configuration()
        self.__load_trigger_data()

    def __load_configuration(self):
        try:
            cwd = os.getcwd()
            config = os.getenv('CONFIG')
            config = (cwd + '/package.json') if config is None else config
            with open(config, "r") as f:
                self._db  = json.load(f)
        except Exception as e:
            self.__load_default()

    def __load_default(self):
        self._db = {}
        self._db['executable'] = {}
        self._db['trigger'] = {}
        self._db['expose'] = {}
        self._db['params'] = {}

    def __load_trigger_data(self):
        data = os.getenv('DATA')
        if data is None:
            self._data = ""
            return

        mode = self._db['trigger'].get('mode', None)
        if mode == 'data':
            self._data = data[5:].strip()

    def data_driven_tags(self):
        trigger = self._db['trigger'].get('dataDriven', None)
        if trigger is None:
            return {}
        return trigger.get('tags', {})

    def data_driven_events(self):
        trigger = self._db['trigger'].get('dataDriven', None)
        if trigger is None:
            return {}
        return trigger.get('events', {})

    def name(self):
        return self._db.get('name', '')

    def expose_tags(self):
        return self._db['expose'].get('tags', [])

    def parameters(self):
        return self._db.get('params', {})

    def trigger_data(self):
        try:
            return json.loads(self._data)
        except Exception as e:
            return ""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: reverseproxy.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='reverseproxy.proto',
  package='trigger',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x12reverseproxy.proto\x12\x07trigger\"+\n\x07Profile\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x02 \x01(\t\"+\n\x07Request\x12\x0f\n\x07headers\x18\x01 \x01(\t\x12\x0f\n\x07payload\x18\x02 \x01(\t\"Z\n\x05Reply\x12!\n\x07profile\x18\x01 \x01(\x0b\x32\x10.trigger.Profile\x12\x0f\n\x07headers\x18\x02 \x01(\t\x12\x0f\n\x07payload\x18\x03 \x01(\t\x12\x0c\n\x04\x63ode\x18\x04 \x01(\x04\x32k\n\x07Trigger\x12\x30\n\x08Register\x12\x10.trigger.Profile\x1a\x10.trigger.Profile\"\x00\x12.\n\x04\x46ire\x12\x0e.trigger.Reply\x1a\x10.trigger.Request\"\x00(\x01\x30\x01\x62\x06proto3'
)




_PROFILE = _descriptor.Descriptor(
  name='Profile',
  full_name='trigger.Profile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='method', full_name='trigger.Profile.method', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='endpoint', full_name='trigger.Profile.endpoint', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=31,
  serialized_end=74,
)


_REQUEST = _descriptor.Descriptor(
  name='Request',
  full_name='trigger.Request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='headers', full_name='trigger.Request.headers', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='trigger.Request.payload', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=76,
  serialized_end=119,
)


_REPLY = _descriptor.Descriptor(
  name='Reply',
  full_name='trigger.Reply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='profile', full_name='trigger.Reply.profile', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='headers', full_name='trigger.Reply.headers', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='trigger.Reply.payload', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='trigger.Reply.code', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=121,
  serialized_end=211,
)

_REPLY.fields_by_name['profile'].message_type = _PROFILE
DESCRIPTOR.message_types_by_name['Profile'] = _PROFILE
DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Reply'] = _REPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Profile = _reflection.GeneratedProtocolMessageType('Profile', (_message.Message,), {
  'DESCRIPTOR' : _PROFILE,
  '__module__' : 'reverseproxy_pb2'
  # @@protoc_insertion_point(class_scope:trigger.Profile)
  })
_sym_db.RegisterMessage(Profile)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), {
  'DESCRIPTOR' : _REQUEST,
  '__module__' : 'reverseproxy_pb2'
  # @@protoc_insertion_point(class_scope:trigger.Request)
  })
_sym_db.RegisterMessage(Request)

Reply = _reflection.GeneratedProtocolMessageType('Reply', (_message.Message,), {
  'DESCRIPTOR' : _REPLY,
  '__module__' : 'reverseproxy_pb2'
  # @@protoc_insertion_point(class_scope:trigger.Reply)
  })
_sym_db.RegisterMessage(Reply)



_TRIGGER = _descriptor.ServiceDescriptor(
  name='Trigger',
  full_name='trigger.Trigger',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=213,
  serialized_end=320,
  methods=[
  _descriptor.MethodDescriptor(
    name='Register',
    full_name='trigger.Trigger.Register',
    index=0,
    containing_service=None,
    input_type=_PROFILE,
    output_type=_PROFILE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Fire',
    full_name='trigger.Trigger.Fire',
    index=1,
    containing_service=None,
    input_type=_REPLY,
    output_type=_REQUEST,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_TRIGGER)

DESCRIPTOR.services_by_name['Trigger'] = _TRIGGER

# @@protoc_insertion_point(module_scope)
